// dx.cpp
// (C) David Joffe 1998
// '98/01

// MFC-only:
#include "stdafx.h"

#include <ddraw.h>
#include <d3drm.h>

#include "dx.h"

BOOL					g_bFullScreen;
BOOL					g_bActive;
BOOL					g_bUse3D;
LPDIRECTDRAW			g_pDD;				// DirectDraw object
LPDIRECTDRAWSURFACE		g_pDDSPrimary;		// Front surface
LPDIRECTDRAWSURFACE		g_pDDSBack;			// Back surface
LPDIRECTDRAWCLIPPER		g_pClipper;			// Clipper
HWND					g_hWnd;				// HWND
SIZE					g_szBack;
RECT					g_rcFront;

LPDIRECT3DRM			g_pD3DRM;			// Direct3DRM object
LPDIRECT3DRMDEVICE		g_pD3DRMDevice;
LPDIRECT3DRMVIEWPORT	g_pViewport;
LPDIRECT3DRMFRAME		g_pCamera;
LPDIRECT3DRMFRAME		g_pScene;


#ifndef TRACEDD
#define TRACEDD(hr, ret) \
if (Failed(hr)) { \
	TRACE("DX Error(%s:%d): %s\n", __FILE__, __LINE__, ErrorToString(hr)); \
	return (ret); \
}
#endif

#ifndef TRACEDX
#define TRACEDX(hr, ret, msg) \
if (Failed(hr, (msg))) { \
	TRACE("Error at (%s:%d)\n", __FILE__, __LINE__); \
	return (ret); \
}
#endif

// Set the pointers to initial values of NULL, etc.
void ClearX()
{
	g_pDD = NULL;
	g_bFullScreen = FALSE;
	g_bActive = FALSE;
	g_bUse3D = FALSE;
	g_pDDSPrimary = NULL;
	g_pDDSBack = NULL;
	g_pClipper = NULL;

	g_pD3DRM = NULL;
	g_pD3DRMDevice = NULL;
	g_pViewport = NULL;
	g_pCamera = NULL;
	g_pScene = NULL;
	
	g_hWnd = 0;
	g_szBack.cx = -1;
	g_szBack.cy = -1;
	SetRect(&g_rcFront, 0, 0, -1, -1);
}

BOOL InitX(HWND hWnd, BOOL bUse3D, BOOL bFullScreen, int iX, int iY, int iDX, int iDY, BOOL bActive)
{
	g_hWnd = hWnd;
	g_bFullScreen = bFullScreen;
	g_bUse3D = bUse3D;
	SETSIZE(&g_szBack, iDX, iDY);
	SetRect(&g_rcFront, iX, iY, iDX, iDY);
	if (Failed(CreateDirectDraw()))
		return FALSE;
	
	if (Failed(DDSetCooperativeLevel(FALSE, g_hWnd)))
		return FALSE;
	
	if (Failed(CreateSurfaces(g_bUse3D, iX, iY, iDX, iDY)))
		return FALSE;
	
	if (Failed(DDSetDisplayMode()))
		return FALSE;

	if (g_bUse3D)
	{
		if (Failed(Init3D()))
			return FALSE;
	}
	
	SetActive(bActive);
	
	return TRUE;
}

HRESULT Init3D()
{
	HRESULT hr;

	CreateDirect3DRM();
	
	CreateDevice();
	
	CreateDefaultScene();
	
	CreateViewport();
	
	return hr;
}

// Create the IDirectDraw object
HRESULT CreateDirectDraw()
{
	HRESULT hr;
	hr = DirectDrawCreate(NULL, &g_pDD, NULL);
	TRACEDD(hr, hr);
	return hr;
}

// Set the cooperativelevel, based on g_bFullScreen
HRESULT DDSetCooperativeLevel(BOOL bFullScreen, HWND hWnd)
{
	HRESULT hr;
	g_bFullScreen = bFullScreen;
	g_hWnd = hWnd;
	if (g_bFullScreen) {
		hr = g_pDD->SetCooperativeLevel(hWnd, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE | DDSCL_ALLOWMODEX | DDSCL_NOWINDOWCHANGES);
	} else {
		hr = g_pDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL);
	}
	TRACEDD(hr, hr);
	return hr;
}

// Create a surface flipping structure
HRESULT CreateSurfaces(BOOL bUse3D, int iX, int iY, int iDX, int iDY)
{
	HRESULT hr;
	DDSURFACEDESC sd;
	memset(&sd, 0, sizeof(sd));
	sd.dwSize = sizeof(sd);
	
	// If we're in fullscreen mode, create front and back together
	if (g_bFullScreen == TRUE)
	{
		sd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		sd.dwBackBufferCount = 1;
		sd.ddsCaps.dwCaps = DDSCAPS_FLIP | DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | (bUse3D ? DDSCAPS_3DDEVICE : 0);
		
		hr = g_pDD->CreateSurface(&sd, &g_pDDSPrimary, NULL);
		TRACEDD(hr, hr);
		
		DDSCAPS ddscaps;
		ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
		hr = g_pDDSPrimary->GetAttachedSurface(&ddscaps, &g_pDDSBack);
		TRACEDD(hr, hr);
	}else{
		// If no width,height specified, use the window size
		if ( (iDX == -1) || (iDY == -1) )
		{
			CRect rc;
			GetClientRect(g_hWnd, &rc);
			iDX = rc.right - rc.left;
			iDY = rc.bottom - rc.top;
		}
		// Create surfaces for windowed mode:
		ASSERT(iDX > 0);
		ASSERT(iDY > 0);
		SETSIZE(&g_szBack, iDX, iDY);
		SetRect(&g_rcFront, iX, iY, iDX, iDY);
		
		// Create the primary surface
		sd.dwFlags = DDSD_CAPS;
		sd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
		hr = g_pDD->CreateSurface(&sd, &g_pDDSPrimary, NULL);
		TRACEDD(hr, hr);
		
		// Create the back surface
		sd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		sd.dwWidth = g_szBack.cx;
		sd.dwHeight = g_szBack.cy;
		sd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | ((bUse3D) ? DDSCAPS_3DDEVICE : 0);
		hr = g_pDD->CreateSurface(&sd, &g_pDDSBack, NULL);
		TRACEDD(hr, hr);
	}
	
	// If we're in a window, create and attach clipper
	if (g_bFullScreen == FALSE)
	{
		if (g_pClipper)
			RELEASE(g_pClipper);
		// Create the clipper
		hr = g_pDD->CreateClipper(0, &g_pClipper, NULL);
		TRACEDD(hr, hr);
		
		// Associate the clipper with the window
		hr = g_pClipper->SetHWnd(0, g_hWnd);
		TRACEDD(hr, hr);
		
		// Attach the clipper to the primary surface (necessary?)
		hr = g_pDDSPrimary->SetClipper(g_pClipper);
		TRACEDD(hr, hr);
	}
	
	return hr;
}

// Set the display mode
HRESULT DDSetDisplayMode(int iWidth, int iHeight, int iBpp)
{
	HRESULT hr;
	if (g_bFullScreen) {
		ASSERT(iWidth > 0);
		ASSERT(iHeight > 0);
		ASSERT(iBpp > 0);
		hr = g_pDD->SetDisplayMode(iWidth, iHeight, iBpp);
		TRACEDD(hr, hr);
		SETSIZE(&g_szBack, iWidth, iHeight);
	}else{
		return DD_OK;
	}
	return hr;
}

// Flip surfaces if full-screen, else blit
HRESULT DDFlip()
{
	HRESULT hr;
	if (g_bActive == FALSE)
		return DD_OK;

	if (g_bFullScreen)
	{
		hr = g_pDDSPrimary->Flip(NULL, DDFLIP_WAIT);
		TRACEDD(hr, hr); // remove for speed? #ifdef DEBUG? Another #DEF?
	}else{
		CRect rcSrc(0, 0, g_szBack.cx, g_szBack.cy);
		CRect rcDest(g_rcFront.left, g_rcFront.top,
			g_rcFront.left + g_rcFront.right - 1,
			g_rcFront.top + g_szBack.cy - 1);
		hr = g_pDDSPrimary->Blt(&rcDest, g_pDDSBack, &rcSrc, DDBLT_WAIT, NULL);
		TRACEDD(hr, hr);
	}

	return hr;
}

// Set whether or not we should be redrawing
void SetActive(BOOL bActive)
{
	g_bActive = bActive;
}

// Determine whether or not we are active
BOOL GetActive()
{
	return g_bActive;
}

// Switch between full-screen and windowed mode. Not working yet.
BOOL SwitchModes()
{
	BOOL bRet = TRUE;
	return bRet;
}

HRESULT DDResize(int iX, int iY, int iDX, int iDY)
{
	HRESULT hr;

	if ( (g_pDDSPrimary == NULL) || (g_pDDSBack == NULL) )
		return DD_OK;
	SetRect(&g_rcFront, iX, iY, iDX, iDY);
	SETSIZE(&g_szBack, iDX, iDY);

	RELEASE(g_pDDSPrimary);
	RELEASE(g_pDDSBack);

	hr = CreateSurfaces(g_bUse3D, iX, iY, iDX, iDY);
	if (Failed(hr))
		return 0;

	if (g_bUse3D)
	{
		CreateDevice();
		
		CreateViewport();
	}

	return DD_OK;

}

BOOL CreateDirect3DRM()
{
	HRESULT hr;

	hr = Direct3DRMCreate(&g_pD3DRM);
	TRACEDX(hr, FALSE, "Error creating Direct3DRM object");

	return TRUE;
}

BOOL CreateDevice()
{
	HRESULT hr;

	if (g_bFullScreen) {
		hr = g_pD3DRM->CreateDeviceFromSurface(NULL, g_pDD, g_pDDSBack, &g_pD3DRMDevice);
		TRACEDX(hr, FALSE, "Error creating device from surface");
	}else{
		hr = g_pD3DRM->CreateDeviceFromSurface(NULL, g_pDD, g_pDDSBack, &g_pD3DRMDevice);
		TRACEDX(hr, FALSE, "Error creating device from surface");
//		hr = g_pD3DRM->CreateDeviceFromClipper(g_pClipper, NULL, g_szBack.cx, g_szBack.cy, &g_pD3DRMDevice);
//		TRACEDX(hr, FALSE, "Error creating device from clipper");
	}
	g_pD3DRMDevice->SetDither(FALSE);

	return TRUE;
}

LPDIRECT3DRMFRAME LoadFrame(char* pszFilename)
{
	HRESULT hr;
	LPDIRECT3DRMFRAME pFrame;

	hr = g_pD3DRM->CreateFrame(g_pScene, &pFrame);
	if (Failed(hr)) {
		TRACE("Error loading frame %s\n");
		return NULL;
	}

	hr = pFrame->Load(pszFilename, NULL, D3DRMLOAD_FROMFILE, loadTextures, NULL);
	if (Failed(hr, "Error loading frame; trying MeshBuilder"))
	{
		LPDIRECT3DRMMESHBUILDER pBuilder;

		if (Failed(g_pD3DRM->CreateMeshBuilder(&pBuilder), "Failed to create meshbuilder"))
			return NULL;

		if (Failed(pBuilder->Load(pszFilename, NULL, D3DRMLOAD_FROMFILE, loadTextures, NULL))) {
			TRACE("(%s): Failed to load object (%s)", pszFilename, ErrorToString(hr));
			return NULL;
		}
		
		if (Failed(pBuilder->SetQuality(D3DRMRENDER_GOURAUD), "Failed to set render quality"))
			return NULL;

		hr = pFrame->AddVisual(pBuilder);
		if (Failed(hr, "Failed to add pBuilder to pFrame"))
			return NULL;
	}

	return pFrame;
/*
	hr = pCar->Load("shut.x", NULL, D3DRMLOAD_FROMFILE, loadTextures, NULL);
	if (Failed(hr)) {
		TRACE("Error [%s] loading \n", ErrorToString(hr));
		return 2;
	}
*/
}

// Put stuff into the scene
// lights (ambient and 1 directional), and a default object
BOOL CreateDefaultScene()
{
	HRESULT hr;

	// Create the scene
	hr = g_pD3DRM->CreateFrame(NULL, &g_pScene);
	if (Failed(hr, "Error creating the scene frame: "))
		return FALSE;

	// Create the camera as a child of the scene
	hr = g_pD3DRM->CreateFrame(g_pScene, &g_pCamera);
	if (Failed(hr, "Error creating the camera: "))
		return FALSE;

//	g_vecPos.x = D3DVAL(0.0);
//	g_vecPos.y = D3DVAL(170.0);
//	g_vecPos.z = D3DVAL(-700.0);
	g_pCamera->SetPosition(g_pScene, D3DVAL(0.0), D3DVAL(0.0), D3DVAL(0.0));
	//pCamera->SetOrientation(pScene, D3DVAL(0.0), D3DVAL(-10.0), D3DVAL(100.0),
	//	 D3DVAL(0.0), D3DVAL(1.0), D3DVAL(0.0));
	// lights? Otherwise we'll just have darkness :-)

		// create lights
	LPDIRECT3DRMLIGHT pLightAmbient = NULL;
	LPDIRECT3DRMLIGHT pLightDirectional = NULL;
	LPDIRECT3DRMFRAME pLights = NULL;

	// Create two lights and a frame to attach them to
	g_pD3DRM->CreateFrame(g_pScene, &pLights);
	g_pD3DRM->CreateLight(D3DRMLIGHT_AMBIENT,     D3DRMCreateColorRGB(D3DVAL(0.3), D3DVAL(0.3), D3DVAL(0.3)), &pLightAmbient);
	g_pD3DRM->CreateLight(D3DRMLIGHT_DIRECTIONAL, D3DRMCreateColorRGB(D3DVAL(0.8), D3DVAL(0.3), D3DVAL(0.3)), &pLightDirectional);

	pLights->AddLight(pLightAmbient);

	pLights->SetPosition(g_pScene, D3DVALUE(0.0), D3DVALUE(0.0), D3DVALUE(0.0));
	pLights->SetOrientation(g_pScene, D3DVALUE(30.0), D3DVALUE(-20.0), D3DVALUE(50.0),
		 D3DVALUE(0.0), D3DVALUE(1.0), D3DVALUE(0.0));

	// Add the lights to the scene
	pLights->AddLight(pLightAmbient);
	pLights->AddLight(pLightDirectional);

	// Release the temporary lights created.
	// It seems they will have been copied for the scene during AddLight
	pLightAmbient->Release();
	pLightDirectional->Release();

	return 0;
}

BOOL CreateViewport(D3DVALUE fBack)
{
	HRESULT hr;

	hr = g_pD3DRM->CreateViewport(g_pD3DRMDevice, g_pCamera, 0, 0, g_szBack.cx, g_szBack.cy, &g_pViewport);
	if (Failed(hr, "Error creating viewport: "))
		return FALSE;

	hr = g_pViewport->SetBack(fBack);
	if (Failed(hr, "Error setting viewport back: "))
		return FALSE;


	return TRUE;
}

// Texture load callback function
HRESULT __cdecl loadTextures(char *name, void *arg, LPDIRECT3DRMTEXTURE *tex)
{
	TRACE("loadTextures(%s)\n", name);
	HRESULT hr;
	hr = g_pD3DRM->LoadTexture(name, tex);
	if (Failed(hr)) {
		TRACE("loadTextures(%s) failed: %s\n", name, ErrorToString(hr));
	}else{
		(*tex)->SetDecalTransparency(TRUE);
		(*tex)->SetDecalTransparentColor(D3DRMCreateColorRGB(255, 0, 255));
	}
	return hr;
}

char* GetFrameName(LPDIRECT3DRMFRAME pFrame)
{
	static char buf[1024];
	DWORD dwSize = (DWORD)sizeof(buf);
	HRESULT hr;
	hr = pFrame->GetName(&dwSize, buf);
	if (Failed(hr))
	{
		TRACE("GetFrameName(): Error %s getting name\n", ErrorToString(hr));
		return NULL;
	}
	return buf;
}

LPDIRECT3DRMFRAME GetFrameWithName(LPDIRECT3DRMFRAME pRootFrame, char* pszName)
{
	HRESULT hr;
	DWORD i;
	char* psz;

	psz = GetFrameName(pRootFrame);

//	TRACE("Testing name: [%s]\n", psz);

	if (psz != NULL)
	{	
		// Found
		if (strcmp(pszName, psz) == 0)
		{
			return pRootFrame;
		}
	}

	LPDIRECT3DRMFRAMEARRAY pChildren;
	hr = pRootFrame->GetChildren(&pChildren);
	if (Failed(hr))
	{
		return NULL;
	}
	for (i = 0; i < pChildren->GetSize(); i++)
	{
		LPDIRECT3DRMFRAME pChildFrame;
		LPDIRECT3DRMFRAME pFrame;
		hr = pChildren->GetElement(i, &pChildFrame);
		if (Failed(hr))
		{
			return NULL;
		}
		pFrame = GetFrameWithName(pChildFrame, pszName);
		if (pFrame != NULL)
		{
			return pFrame;
		}
	}
	
	return NULL;
}

/*
int SetQuality(LPDIRECT3DRMFRAME pFrame, D3DRMRENDERQUALITY quality)
{
	HRESULT hr;
	LPDIRECT3DRMMESHBUILDER pVisual;
//	hr = pD3DRM->GetNamedObject("cutters", (LPDIRECT3DRMOBJECT*)&pVisual);
	//if (Failed(hr))
	//{
		//TRACE("dskfjssdlf\n");
		//return -1;
	//}
	//pVisual->SetQuality(quality);
	//return 0;
	DWORD dw;
	LPDIRECT3DRMVISUALARRAY pVisuals;
	hr = pFrame->GetVisuals(&pVisuals);
	if (Failed(hr))
	{
		TRACE("dskfjssdsfddsfsdfdsflf\n");
		return -1;
	}
	TRACE("\nNumber of visuals: %d\n", (int)pVisuals->GetSize());
	for (dw = 0; dw < pVisuals->GetSize(); dw++)
	{
		hr = pVisuals->GetElement(dw, (LPDIRECT3DRMVISUAL*)&pVisual);
		if (Failed(hr))
		{
			TRACE("ksdjfklhkjsdahfkljasdhfkjsdahads\n");
			return -2;
		}
		pVisual->SetQuality(quality);
	}
	return 0;
}
*/

/*
HRESULT PASCAL EnumerateDriversCallback(LPGUID lpGuid,
LPSTR lpDeviceDescription,
LPSTR lpDeviceName,
	LPD3DDEVICEDESC lpD3DHWDeviceDesc,
	LPD3DDEVICEDESC lpD3DHELDeviceDesc,
	LPVOID lpUserArg 
)
{
	CD3DDriver* pDriver;

	TRACE("Driver: [%s] (%s)\n", lpDeviceName, lpDeviceDescription);

	pDriver = new CD3DDriver(lpDeviceName, lpDeviceDescription);
	apDrivers.Add(pDriver);

	return D3DENUMRET_OK;
}

HRESULT EnumerateDrivers()
{
	HRESULT hr;
	hr = pD3D->EnumDevices(EnumerateDriversCallback, NULL);
	if (Failed(hr))
	{
		TRACE("lksdfjlaj;lkdsladsf\n");
		return hr;
	}
	return hr;
}

*/